LOGIN_ENDPOINT = "/WManage/api/login"
INVERTER_RUNTIME_ENDPOINT = "/WManage/api/inverter/getInverterRuntime"
INVERTER_ENERGY_ENDPOINT = "/WManage/api/inverter/getInverterEnergyInfo"
INVERTER_BATTERY_ENDPOINT = "/WManage/api/battery/getBatteryInfo"
INVERTER_PARAMETER_READ = "/WManage/web/maintain/remoteRead/read"
INVERTER_PARAMETER_WRITE = "/WManage/web/maintain/remoteSet/write"



if __name__ == "__main__":
    import sys
    import os
    sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../..")))

import asyncio
import logging

import aiohttp

from eg4_inverter_api.exceptions import EG4APIError, EG4AuthError

from eg4_inverter_api.constants import (
    INVERTER_BATTERY_ENDPOINT,
    INVERTER_ENERGY_ENDPOINT,
    INVERTER_RUNTIME_ENDPOINT,
    LOGIN_ENDPOINT,
    INVERTER_PARAMETER_READ,
    INVERTER_PARAMETER_WRITE,
)

from eg4_inverter_api.models import (
    APIResponse,
    BatteryData,
    BatteryUnit,
    EnergyData,
    Inverter,
    RuntimeData,
    InverterParameters,
)


class EG4InverterAPI:
    """Asynchronous EG4 API client."""

    def __init__(
        self, username, password, serialNum=None, plantId=None, base_url=None
    ) -> None:
        self._username = username
        self._password = password
        self._session = None
        self.jsessionid = None
        self._inverters = []
        self._plantId = plantId
        self._serialNum = serialNum
        self._base_url = base_url or "https://monitor.eg4electronics.com"
        self._ignore_ssl = False

        # Endpoint references
        self._login_url = f"{self._base_url}{LOGIN_ENDPOINT}"
        self._inverter_runtime_url = f"{self._base_url}{INVERTER_RUNTIME_ENDPOINT}"
        self._inverter_energy_url = f"{self._base_url}{INVERTER_ENERGY_ENDPOINT}"
        self._inverter_battery_url = f"{self._base_url}{INVERTER_BATTERY_ENDPOINT}"
        self._inverter_parameter_read = f"{self._base_url}{INVERTER_PARAMETER_READ}"
        self._inverter_parameter_write = f"{self._base_url}{INVERTER_PARAMETER_WRITE}"

    async def _get_session(self):
        """Initialize an aiohttp session."""
        headers={
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        do_ssl = not self._ignore_ssl

        if self._session is None or self._session.closed:
            self._session = aiohttp.ClientSession(
                connector=aiohttp.TCPConnector(ssl=do_ssl),
                headers=headers
            )
        return self._session

    async def login(self, ignore_ssl=False) -> None:
        """Authenticate and retrieve session cookie."""
        self._ignore_ssl = ignore_ssl
        session = await self._get_session()

        payload = f"account={self._username}&password={self._password}"
        async with session.post(self._login_url, data=payload) as response:
            if response.status == 200:
                self.jsessionid = self._extract_jsessionid(response.cookies)
                inverter_data = await response.json()
                if "success" not in inverter_data or inverter_data["success"] is not True:
                    raise EG4AuthError("Login failed. Please check your credentials.")
                self._inverters = self._extract_inverters(inverter_data)
                logging.info(f"Login successful for user {self._username}")
            else:
                raise EG4AuthError("Login failed. Please check your credentials.")

    def _extract_jsessionid(self, cookies):
        """Extract JSESSIONID from cookies."""
        for key, cookie in cookies.items():
            if key == "JSESSIONID":
                return cookie.value
        raise EG4AuthError("Failed to retrieve JSESSIONID during login.")

    def _extract_inverters(self, data):
        """Extract available inverters from the login response."""
        inverters = []

        for plant in data.get("plants", []):
            plantId = plant.get("plantId")
            plant_name = plant.get("name")
            for inverter in plant.get("inverters", []):
                inverter = Inverter(
                    plantId=plantId,
                    plantName=plant_name,
                    captureExtra=True,
                    **inverter,
                )
                inverters.append(inverter)

        if not inverters:
            raise EG4APIError("No inverters found in the response data.")

        return inverters

    async def _request(self, method, url, payload=None):
        """Unified async request method with automatic reauthentication."""
        if not self.jsessionid:
            await self.login()

        session = await self._get_session()

        headers = {"Cookie": f"JSESSIONID={self.jsessionid}"}
        async with session.request(
            method, url, headers=headers, data=payload
        ) as response:
            if response.status == 401:
                await self.login()  # Re-authenticate on 401
                async with session.request(
                    method, url, headers=headers, data=payload
                ) as retry_response:
                    if retry_response.status != 200:
                        raise EG4APIError(
                            f"API request failed: {retry_response.status}"
                        )
                    return await retry_response.json()

            if response.status != 200:
                raise EG4APIError(
                    f"API request failed: {response.status} - {await response.text()}"
                )

            return await response.json()

    async def get_inverter_runtime_async(self, captureExtra=True):
        """Retrieve inverter runtime data."""
        payload = f"serialNum={self._serialNum}"
        response = await self._request("POST", self._inverter_runtime_url, payload)

        if response.get("success"):
            return RuntimeData(captureExtra=captureExtra, **response)
        else:
            return APIResponse(success=False, error_message=response.get("error"))

    async def get_inverter_energy_async(self, captureExtra=True):
        """Retrieve inverter energy data."""
        payload = f"serialNum={self._serialNum}"
        response = await self._request("POST", self._inverter_energy_url, payload)
        if response.get("success"):
            return EnergyData(captureExtra=captureExtra, **response)
        else:
            return APIResponse(success=False, error_message=response.get("error"))

    async def get_inverter_battery_async(self, captureExtra=True):
        """Retrieve inverter battery data."""
        payload = f"serialNum={self._serialNum}"
        response = await self._request("POST", self._inverter_battery_url, payload)

        if response.get("success"):
            # Extract battery units
            battery_units = [
                BatteryUnit(captureExtra=captureExtra, **unit)
                for unit in response.get("batteryArray", [])
            ]

            # Extract overall battery data
            return BatteryData(
                remainCapacity=response.get("remainCapacity"),
                fullCapacity=response.get("fullCapacity"),
                totalNumber=response.get("totalNumber"),
                totalVoltageText=response.get("totalVoltageText"),
                currentText=response.get("currentText"),
                battery_units=battery_units,
            )
        else:
            return APIResponse(success=False, error_message=response.get("error"))

    async def read_settings_async(self):
        """Read inverter settings across 3 register ranges."""
        inverterParameters = InverterParameters()
        success = True
        
        for start_register in [0, 127, 240, 500, 2000, 5000]:
            payload = f"inverterSn={self._serialNum}&startRegister={start_register}&pointNumber=127&autoRetry=true"
            response = await self._request("POST", self._inverter_parameter_read, payload)
            if response.get("success"):
                inverterParameters.from_dict(response)
            else:
                return APIResponse(success=False, error_message=response.get("error"))

        return inverterParameters

    async def write_setting_async(self, hold_param, value_text):
        """Write a single inverter setting."""
        payload = f"inverterSn={self._serialNum}&holdParam={hold_param}&valueText={value_text}&clientType=WEB&remoteSetType=NORMAL"
        response = await self._request("POST", self._inverter_parameter_write, payload)
        return response.get("success")
    
    async def close(self):
        """Close the aiohttp session when done."""
        if self._session:
            await self._session.close()

    # --------- SYNC WRAPPERS ---------
    def get_inverter_runtime(self, captureExtra=True):
        """Sync wrapper for inverter runtime data."""
        return asyncio.run(self.get_inverter_runtime_async(captureExtra))

    def get_inverter_energy(self, captureExtra=True):
        """Sync wrapper for inverter energy data."""
        return asyncio.run(self.get_inverter_energy_async(captureExtra))
    
    def get_inverter_battery(self, captureExtra=True):
        """Sync wrapper for inverter battery data."""
        return asyncio.run(self.get_inverter_battery_async(captureExtra))

    def read_settings(self):
        """Sync wrapper for inverter battery data."""
        return asyncio.run(self.read_settings_async())

    def write_settings(self, hold_param, value_text):
        """Sync wrapper for inverter battery data."""
        return asyncio.run(self.write_setting_async(hold_param, value_text))

    # --------- SYNC FUNCTIONS  ---------
    def get_inverters(self):
        """Show the list of inverters."""
        return self._inverters

    def set_selected_inverter(
        self, plantId=None, serialNum=None, inverterIndex=None
    ) -> None:
        """Set the plantId and serialNum."""
        # TODO: handle serialNum without plant (discover plant)
        if serialNum is not None:
            self._serialNum = serialNum
            inverter = [ x for x in self._inverters if x.serialNum == serialNum ]
            if len(inverter) == 0:
                inverter = inverter[0]
                self._plantId = inverter.plantId
            else:
                inverter = None
        elif inverterIndex is not None and len(self._inverters) > inverterIndex:
            inverter = self._inverters[inverterIndex]
            self._plantId = inverter.plantId
            self._serialNum = inverter.serialNum
        else:
            raise EG4APIError("No Inverter or Plant/Serial selection")

    def get_selecter_inverter(self) -> Inverter:
        """ retrieve the selected inverter """
        inverter = [ x for x in self._inverters if x.serialNum == self._serialNum ]
        if len(inverter) == 0:
            return inverter[0]
        
        return None
        
# --------- MAIN EXECUTION BLOCK FOR TESTING ---------
if __name__ == "__main__":
    logging.basicConfig(
        level=logging.INFO,  # Show INFO and above
        format="%(asctime)s - %(levelname)s - %(message)s",  # Clear timestamped format
    )
    import os
    from dotenv import load_dotenv

    load_dotenv()

    async def test():
        USERNAME = os.getenv("EG4_USERNAME")
        PASSWORD = os.getenv("EG4_PASSWORD")
        SERIAL_NUMBER = os.getenv("EG4_SERIAL_NUMBER")
        PLANT_ID = os.getenv("EG4_PLANT_ID")
        BASE_URL = os.getenv("EG4_BASE_URL", "https://monitor.eg4electronics.com")
        IGNORE_SSL = os.getenv("EG4_DISABLE_VERIFY_SSL", "0") == "1"
        
        # api = EG4InverterAPI(USERNAME, PASSWORD, serialNum, plantId)
        api = EG4InverterAPI(USERNAME, PASSWORD, BASE_URL)

        try:
            await api.login(ignore_ssl=IGNORE_SSL)  # noqa: SLF001
            inverters = api.get_inverters()
            selected_inverter = [
                x for x in inverters if x.serialNum== "43730P0090"
            ]
            
            api.set_selected_inverter(inverterIndex=0)

            # Show available inverters
            logging.info("Inverters Found:")
            for index, inverter in enumerate(api.get_inverters()):
                logging.info(f"  {index} - {inverter}")  # noqa: G004

            # Runtime Data
            runtime_data = await api.get_inverter_runtime_async()
            logging.info(f"\nInverter Runtime Data:\n{runtime_data}")

            # Energy Data
            energy_data = await api.get_inverter_energy_async()
            logging.info(f"\nInverter Energy Data:\n{energy_data}")

            # Battery Data
            battery_data = await api.get_inverter_battery_async()
            logging.info(f"\nBattery Data:\n{battery_data}")
            for unit in battery_data.battery_units:
                logging.info(f"\t{unit}")
                
            # Read Parameters
            params = await api.read_settings_async()
            logging.info(f"\nparameters:\n{params}")
            
        except EG4AuthError as auth_err:
            logging.error(f"❌ Authentication Error: {auth_err}")
        except EG4APIError as api_err:
            logging.error(f"❌ API Error: {api_err}")
        except Exception as e:
            logging.error(f"❌ Unexpected Error: {e}")
        finally:
            await api.close()

    asyncio.run(test())



import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '../src')))
import pytest
from dotenv import load_dotenv
from eg4_inverter_api import EG4InverterAPI, EG4AuthError, EG4APIError

# Load environment variables for testing
load_dotenv(".env")

USERNAME = os.getenv("EG4_USERNAME")
PASSWORD = os.getenv("EG4_PASSWORD")
SERIAL_NUMBER = os.getenv("EG4_SERIAL_NUMBER")
PLANT_ID = os.getenv("EG4_PLANT_ID")
BASE_URL = os.getenv("EG4_BASE_URL", "https://monitor.eg4electronics.com")
IGNORE_SSL = os.getenv("EG4_DISABLE_VERIFY_SSL", "1") == "1"

@pytest.mark.asyncio
async def test_login():
    """Test successful login and exercise the functions"""
    api = EG4InverterAPI(USERNAME, PASSWORD, BASE_URL)
    print(api._password)
    await api.login(ignore_ssl=IGNORE_SSL)
    api.set_selected_inverter(inverterIndex=0)
    assert api.jsessionid is not None
    print("Logged in")
    data = await api.get_inverter_runtime_async()
    assert data.success
    assert data.statusText is not None
    print("get_inverter_runtime_async success")
    data = await api.get_inverter_energy_async()
    assert data.success
    print("get_inverter_energy_async success")
    data = await api.get_inverter_battery_async()
    assert data.remainCapacity is not None
    print("get_inverter_battery_async success")
    inverter_params = await api.read_settings_async()
    assert inverter_params.success is True
    assert hasattr(inverter_params, 'inverterRuntimeDeviceTime')
    print("read_settings_async success")
    attr = "OFF_GRID_HOLD_GEN_CHG_START_SOC" if hasattr(inverter_params, "OFF_GRID_HOLD_GEN_CHG_START_SOC") else "HOLD_BATTERY_WARNING_SOC"
    orig_value = getattr(inverter_params,attr)
    assert orig_value is not None
    new_value = int(orig_value) - 1
    data = await api.write_setting_async(hold_param=attr, value_text=new_value)
    assert data is True
    data = await api.write_setting_async(hold_param=attr, value_text=orig_value)
    print("write_setting_async success")
    
@pytest.mark.asyncio
async def test_invalid_login():
    """Test handling of invalid credentials."""
    api2 = EG4InverterAPI(USERNAME, "xxx", BASE_URL)
    with pytest.raises(EG4AuthError):
        await api2.login(ignore_ssl=IGNORE_SSL)



class Inverter:
    """Represents an EG4 Inverter."""

    def __init__(self, plantId, plantName, captureExtra=True, **kwargs) -> None:
        self.plantId = plantId
        self.plantName = plantName
        self.serialNum = None
        self._main_args = [
            "serialNum",
            "phase",
            "dtc",
            "deviceType",
            "subDeviceType",
            "allowExport2Grid",
            "batteryType",
            "standard",
            "slaveVersion",
            "fwVersion",
            "allowGenExercise",
            "withbatteryData",
            "hardwareVersion",
            "voltClass",
            "machineType",
            "protocolVersion",
        ]
        for key in self._main_args:
            setattr(self, key, kwargs.get(key))

        if captureExtra:
            self.from_dict(
                {x: y for x, y in kwargs.items() if x not in self._main_args}
            )

    def from_dict(self, d):
        """Set values based on dictionary."""
        for key, value in d.items():
            setattr(self, key, value)

    def __repr__(self):
        return (
            "Inverter("
            f"serialNum={self.serialNum}, plantName={self.plantName} "
            f"plantName={self.plantId}, batteryType={self.batteryType}, "
            f"fwVersion={self.fwVersion}, phase={self.phase}"
            ")"
        )


class BatteryUnit:
    """Represents an individual battery unit."""

    def __init__(self, captureExtra=True, **kwargs):
        """Initialize BatteryUnit."""
        self._main_args = [
            "batteryKey",
            "batIndex",
            "batterySn",
            "totalVoltage",
            "current",
            "soc",
            "soh",
            "cycleCnt",
        ]
        for key in self._main_args:
            setattr(self, key, kwargs.get(key))

        # Capture any unknown or new API fields dynamically
        if captureExtra:
            self.from_dict(
                {x: y for x, y in kwargs.items() if x not in self._main_args}
            )

    def from_dict(self, d) -> None:
        """Set values based on dictionary."""
        for key, value in d.items():
            setattr(self, key, value)

    def to_dict(self):
        """Return the battery unit as a dictionary."""
        return self.__dict__

    def __repr__(self):
        return f"BatteryUnit({self.batIndex}: sn {self.batterySn}, soc% {self.soc}, soh% {self.soh}, cycles {self.cycleCnt}, v {self.totalVoltage / 100}, )"


class BatteryData:
    """Represents overall battery data including individual battery units."""

    def __init__(
        self,
        remainCapacity,
        fullCapacity,
        totalNumber,
        totalVoltageText,
        currentText,
        battery_units=[],
    ):
        self.remainCapacity = remainCapacity
        self.fullCapacity = fullCapacity
        self.totalNumber = totalNumber
        self.totalVoltageText = totalVoltageText
        self.currentText = currentText

        # Battery units as a list
        self.battery_units = battery_units if battery_units else []

    def to_dict(self):
        """Return the battery data including individual units as a dictionary."""
        data = self.__dict__.copy()
        data["battery_units"] = [unit.to_dict() for unit in self.battery_units]
        return data

    def __repr__(self):
        d = self.to_dict()
        data = {x: y for x, y in d.items() if x != "battery_units"}
        return f"BatteryData({data.items()} )"


class EnergyData:
    """Represents inverter energy data from the API."""

    def __init__(self, captureExtra=True, **kwargs):
        self._main_args = [
            "todayYielding",
            "totalYielding",
            "todayDischarging",
            "totalDischarging",
            "todayCharging",
            "totalCharging",
            "todayImport",
            "totalImport",
            "todayExport",
            "totalExport",
            "todayUsage",
            "totalUsage",
        ]
        for key in self._main_args:
            setattr(self, key, kwargs.get(key))

        # Capture any unknown or new API fields dynamically
        if captureExtra:
            self.from_dict(
                {x: y for x, y in kwargs.items() if x not in self._main_args}
            )

    def from_dict(self, d) -> None:
        """Set values based on dictionary."""
        for key, value in d.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"EnergyData({self.to_dict()})"

    def to_dict(self):
        return self.__dict__


class RuntimeData:
    """Represents inverter runtime data from the API."""

    def __init__(
        self,
        captureExtra=True,
        **kwargs,
    ):
        self._main_args = [
            "statusText",
            "batteryType",
            "batParallelNum",
            "batCapacity",
            "consumptionPower",
            "vpv1",
            "vpv2",
            "vpv3",
            "vpv4",
            "ppvpCharge",
            "pDisCharge",
            "peps",
            "pToGrid",
            "pToUser",
        ]
        for key in self._main_args:
            setattr(self, key, kwargs.get(key))

        # Capture any unknown or new API fields dynamically
        if captureExtra:
            d = {x: y for x, y in kwargs.items() if x not in self._main_args}
            self.from_dict(d)

    def from_dict(self, d) -> None:
        """Set values based on dictionary."""
        for key, value in d.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"RuntimeData({self.to_dict()})"

    def to_dict(self):
        return self.__dict__

class InverterParameters:
    """Represents inverter parameters."""

    def __init__(
        self
    ):
        self._skip_args = ["valueFrame","inverterSn","startRegister","pointNumber"]
        self._main_args = ["success"]
        self.success=True

    def from_dict(self, d) -> None:
        """Set values based on dictionary."""
        for key, value in d.items():
            if key not in self._main_args and key not in self._skip_args:
                setattr(self, key, value)

    def __repr__(self):
        return f"RuntimeData({self.to_dict()})"

    def to_dict(self):
        return self.__dict__


class APIResponse:
    """A general-purpose response model for handling success/failure states."""

    def __init__(self, success, data=None, error_message=None):
        self.success = success
        self.data = data
        self.error_message = error_message

    def __repr__(self):
        return f"APIResponse(success={self.success}, data={self.data}, error={self.error_message})"



from setuptools import setup, find_packages

setup(
    name="eg4_inverter_api",
    version="0.1.4",
    description="A Python API for interacting with EG4 Inverter systems.",
    author="Garreth Jeremiah",
    author_email="twistedroutes76@example.com",
    url="https://github.com/twistedroutes/eg4_inverter_api",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: APACHE License",
    ],
    python_requires=">=3.8",
)